from dataclasses import dataclass

from helpers import create_request, prepare_model
from text_generation_server.auto_generator import AutoGenerator
from text_generation_server.pb.generate_pb2 import Batch
from tqdm import tqdm


@dataclass
class DecodeTestParams:
    model_id: str
    sequence_length: int
    expected_text: str
    do_sample: bool = False
    max_new_tokens: int = 20
    top_k: int = 50


def decode_single_test(params):
    model_path = prepare_model(params.model_id, params.sequence_length)
    input_text = "It was a bright cold day in April, and the clocks were striking thirteen."
    max_new_tokens = params.max_new_tokens

    generator = AutoGenerator.from_pretrained(
        model_path, revision="", max_batch_size=1, max_sequence_length=params.sequence_length
    )
    request = create_request(
        id=0,
        inputs=input_text,
        max_new_tokens=max_new_tokens,
        do_sample=params.do_sample,
        top_k=params.top_k,
    )
    batch = Batch(id=0, requests=[request], size=1, max_tokens=params.sequence_length)
    generations, next_batch = generator.prefill(batch)
    # We already generated one token: call decode max_new_tokens - 1 times
    for _ in tqdm(range(max_new_tokens - 1)):
        assert next_batch.size == 1
        assert next_batch.max_tokens == params.sequence_length
        assert len(generations) == 1
        assert len(generations[0].tokens.ids) == 1
        generations, next_batch = generator.decode([next_batch])
    # Destroy generator: this will properly stop threads and prevent them from getting stuck if one of the following
    # assertions fails.
    del generator
    assert next_batch is None
    assert len(generations) == 1
    output = generations[0].generated_text
    assert output.generated_tokens == max_new_tokens
    assert output.finish_reason == 0
    print(f"Generated text: {output.text}")
    if params.do_sample:
        assert output.text != params.expected_text
    else:
        assert output.text == params.expected_text




from jetstream_pt.third_party.mixtral import config as mixtral_config
from jetstream_pt.third_party.mixtral.model import Transformer
from transformers import GenerationConfig, GenerationMixin, MixtralConfig


class MixtralModelHf(Transformer, GenerationMixin):
    """Transformer module that uses HF MixtralConfig instead of Jetstream Pytorch MixtralConfig + device.
    """

    def __init__(
        self,
        config: MixtralConfig,
        device,
        env,
    ):
        self.config = config
        self.generation_config = GenerationConfig.from_model_config(config)

        args = mixtral_config.ModelArgs(
            block_size=config.max_position_embeddings,
            vocab_size=config.vocab_size,
            n_layer=config.num_hidden_layers,
            n_head=config.num_attention_heads,
            dim=config.hidden_size,
            intermediate_size=config.intermediate_size,
            n_local_heads=config.num_local_experts or config.num_attention_heads,
            num_activated_experts=config.num_experts_per_tok,
            device=device,
        )
        super().__init__(args, env)


    @classmethod
    def from_config(cls, config, env):
        device = "meta"
        model = cls(config, device, env)
        return model



from helpers import create_request, prepare_model
from text_generation_server.auto_generator import AutoGenerator
from text_generation_server.pb.generate_pb2 import Batch


def test_prefill_truncate():
    model_id="Maykeye/TinyLLama-v0"
    sequence_length=1024

    model_path = prepare_model(model_id, sequence_length)
    max_new_tokens = 20

    generator = AutoGenerator.from_pretrained(
        model_path, revision="", max_batch_size=1, max_sequence_length=sequence_length
    )
    input_text = "This is something I will tell by the end of the story"

    request = create_request(id=0, inputs=input_text, max_new_tokens=max_new_tokens, do_sample=False)
    batch = Batch(id=0, requests=[request], size=1, max_tokens=sequence_length)
    generations, _ = generator.prefill(batch)
    assert len(generations) == 1
    assert generations[0].tokens.ids == [31843]
    assert generations[0].tokens.texts == ["."]

    # Now re-test but with truncate
    generator.clear()

    request = create_request(id=0, inputs=input_text, max_new_tokens=max_new_tokens, do_sample=False)
    # This will only leave last tokens
    request.truncate = 6
    batch = Batch(id=0, requests=[request], size=1, max_tokens=sequence_length)
    generations, _ = generator.prefill(batch)
    assert len(generations) == 1
    assert generations[0].tokens.ids == [291]
    assert generations[0].tokens.texts == [" and"]



import os

import pytest
import torch
from transformers import AutoTokenizer

from optimum.tpu.distributed_model import DistributedModel


def sample_greedy(logits):
    next_logits = logits[:, -1]
    next_token_id = torch.argmax(next_logits, dim=-1)[:, None].int()
    return next_token_id


def _test_distributed_model_prefill(model_id):
    # This test ensures model can be loaded in a parallel way and
    # that the "proxy" distributed model can be used to prefill the model.
    # Disable tokenizers parallelism to avoid deadlocks
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    tokenizer = AutoTokenizer.from_pretrained(model_id)
    text = ["Running something in parallel means"]
    inputs = tokenizer(text, return_tensors="pt")
    input_ids = inputs["input_ids"]
    attention_mask = inputs["attention_mask"]
    pos_ids = (attention_mask.cumsum(-1) - 1).masked_fill(attention_mask == 0, 0)
    tokens = input_ids.clone()

    model = DistributedModel(model_id, sample_greedy)
    next_tokens = model.prefill(**inputs, position_ids=pos_ids)
    tokens = torch.cat([tokens, next_tokens], dim=-1)

    # Data can be decoded even before leaving
    decoded_texts = tokenizer.batch_decode(tokens, skip_special_tokens=True)
    print()
    print("------------------------------------------")
    print("Decoded texts:")
    print(decoded_texts[0])
    print("------------------------------------------")
    # Even if models are different, for this simple test results are the same.
    expected_text = "Running something in parallel means that"
    assert expected_text == decoded_texts[0]


def test_distributed_model_prefill_gpt2():
    _test_distributed_model_prefill("openai-community/gpt2")


@pytest.mark.slow
def test_distributed_model_prefill_gemma7b():
    _test_distributed_model_prefill("google/gemma-7b")

@pytest.mark.slow
def test_distributed_model_prefill_llama3_8b():
    _test_distributed_model_prefill("meta-llama/Meta-Llama-3-8B")




import pytest
from helpers import create_request, prepare_model
from text_generation_server.auto_generator import AutoGenerator
from text_generation_server.pb.generate_pb2 import Batch
from tqdm import tqdm


MODEL_ID = "openai-community/gpt2"
SEQUENCE_LENGTH = 1024


@pytest.fixture(scope="module")
def model_path():
    return prepare_model(MODEL_ID, SEQUENCE_LENGTH)


def test_info(model_path):
    generator = AutoGenerator.from_pretrained(model_path, revision="", max_batch_size=1, max_sequence_length=1)
    info = generator.info
    assert info.requires_padding is True
    assert info.device_type == "xla"
    assert info.window_size == 0
    assert info.speculate == 0


@pytest.mark.parametrize(
    "input_text, token_id, token_text, do_sample",
    [
        [
            "It was a bright cold day in April, and the clocks were striking thirteen.",
            383,
            " The",
            False,
        ],
        [
            "It was a bright cold day in April, and the clocks were striking thirteen.",
            775,
            " We",
            True,
        ],
    ],
    ids=["greedy", "sample"],
)
@pytest.mark.parametrize("batch_size", [1, 4], ids=["single", "multiple"])
def test_prefill(input_text, token_id, token_text, do_sample, batch_size, model_path):
    generator = AutoGenerator.from_pretrained(model_path, revision="", max_batch_size=batch_size, max_sequence_length=SEQUENCE_LENGTH)
    requests = []
    max_new_tokens = 20
    for i in range(batch_size):
        requests.append(create_request(id=0, inputs=input_text, do_sample=do_sample, max_new_tokens=max_new_tokens))
    # Let's be pessimistic when estimating max_tokens
    batch_size * (len(input_text) + max_new_tokens)
    batch = Batch(id=0, requests=requests, size=batch_size, max_tokens=batch_size * SEQUENCE_LENGTH)
    generations, next_batch = generator.prefill(batch)
    assert next_batch.size == batch_size
    # Whatever was passed as max_tokens, the server will correct it
    # because of static batching
    assert next_batch.max_tokens == batch_size * SEQUENCE_LENGTH
    assert len(generations) == batch_size
    for g in generations:
        tokens = g.tokens
        assert tokens.ids == [token_id]
        assert tokens.texts == [token_text]


def test_decode_multiple(model_path):
    generator = AutoGenerator.from_pretrained(model_path,
                                             revision="",
                                             max_batch_size=2,
                                             max_sequence_length=SEQUENCE_LENGTH)
    input_text = "Once upon a time"
    max_new_tokens = 20
    # Prefill a single request, remembering the generated token
    tokens = {0: [], 1: []}
    request = create_request(id=0, inputs=input_text, max_new_tokens=max_new_tokens)
    batch = Batch(id=0, requests=[request], size=1, max_tokens=SEQUENCE_LENGTH)
    generations, next_batch = generator.prefill(batch)
    assert next_batch.size == 1
    assert len(generations) == 1
    g = generations[0]
    tokens[g.request_id].append(g.tokens.ids[0])
    assert len(tokens[0]) == 1
    # Decode a few tokens
    gen_tokens = 4
    for _ in tqdm(range(gen_tokens - 1), "Decoding tokens"):
        generations, next_batch = generator.decode([next_batch])
        assert len(generations) == 1
        g = generations[0]
        tokens[g.request_id].append(g.tokens.ids[0])
    assert len(tokens[0]) == gen_tokens
    assert next_batch.size == 1
    # Add a second request
    request = create_request(id=1, inputs=input_text, max_new_tokens=max_new_tokens)
    batch = Batch(id=1, requests=[request], size=1, max_tokens=SEQUENCE_LENGTH)
    generations, next_batch_1 = generator.prefill(batch)
    assert next_batch_1.size == 1
    # We should have generated only a single token
    assert len(generations) == 1
    g = generations[0]
    tokens[g.request_id].append(g.tokens.ids[0])
    assert len(tokens[0]) == gen_tokens
    assert len(tokens[1]) == 1
    # Decode more tokens until we reach the maximum for the first request
    batches = [next_batch, next_batch_1]
    for _ in tqdm(range(max_new_tokens - gen_tokens), "Decoding tokens (2nd batch)"):
        generations, next_batch = generator.decode(batches)
        for g in generations:
            tokens[g.request_id].append(g.tokens.ids[0])
        batches = [next_batch]
    # Verify we now only have one pending request
    assert next_batch.size == 1
    assert len(tokens[0]) == max_new_tokens
    assert len(tokens[1]) == max_new_tokens - gen_tokens + 1
    # Verify we have the output for the first request
    for g in generations:
        if g.request_id == 0:
            output = g.generated_text
            assert output.text != ""
            assert output.generated_tokens == max_new_tokens
            generated_text = output.text
    # Continue decoding until the end of the second request
    for _ in tqdm(range(gen_tokens - 1), "Decoding tokens (finishing)"):
        generations, next_batch = generator.decode([next_batch])
        assert len(generations) == 1
        g = generations[0]
        tokens[g.request_id].append(g.tokens.ids[0])
    assert next_batch is None
    output = generations[0].generated_text
    assert output.generated_tokens == max_new_tokens
    assert tokens[0] == tokens[1]
    assert output.text == generated_text



# Import torch_xla2 first
import torch_xla2  # isort:skip

import os
import time
from typing import TYPE_CHECKING, Any

import jax
import torch
from jetstream_pt import fetch_models, quantize_model, torchjax
from jetstream_pt.engine import PyTorchEngine
from jetstream_pt.environment import (
    JetEngineEnvironment,
    JetEngineEnvironmentData,
    QuantizationConfig,
)
from loguru import logger


if TYPE_CHECKING:
    from transformers import PretrainedConfig

from transformers import AutoConfig

from .compatibility import model_can_use_jetstream_pt
from .models import GemmaModel, LlamaModel, MixtralModel


class OptimumJetstreamEngine(PyTorchEngine):
    """This is essentially the same as the PytorchEngine, but it also supports a callback for sampling in prefill and
    generation that can change on each request while not needing to be JIT'ed.
    """
    prefill_ex = PyTorchEngine.prefill

    def __init__(
        self,
        pt_model: torch.nn.Module,
        env: JetEngineEnvironment,
        weights=None,
    ):
        super().__init__(pt_model, env, weights)
        # Call model prefill and generate needs to be JIT'ed, because it is called with sharded notations, and it would
        # otherwise not work for some models.
        self._call_model_prefill = jax.jit(
            self._call_model_prefill,
        )
        self._call_model_generate = jax.jit(
            self._call_model_generate,
        )

def _get_head_dim(config: "PretrainedConfig") -> int:
    if hasattr(config, "head_dim"):
        return config.head_dim
    return config.hidden_size // config.num_attention_heads

def load_model_info(config: "PretrainedConfig") -> Any:
    num_layers = config.num_hidden_layers
    num_heads = config.num_attention_heads
    head_dim = _get_head_dim(config)
    num_kv_heads = config.num_key_value_heads
    n_reps = num_heads // num_kv_heads
    if config.model_type == "llama":
        model_class = LlamaModel
    elif config.model_type == "gemma":
        model_class = GemmaModel
    elif config.model_type == "mixtral":
        model_class = MixtralModel
    else:
        raise ValueError(f"Unsupported model type {config.model_type}")
    model_info = fetch_models.ModelInfo(
        model_class=model_class,
        num_layers=num_layers,
        num_kv_heads=num_kv_heads,
        head_dim=head_dim,
        n_reps=n_reps,
    )
    return model_info


def create_engine_env_data(
    model_path: str,
    batch_size: int,
    sequence_length: int,
    max_input_tokens: int,
    max_output_tokens: int,
) -> Any:
    if not model_can_use_jetstream_pt(model_path):
        return None
    # First get config
    config = AutoConfig.from_pretrained(model_path)
    model_info = load_model_info(config)
    if model_info is None:
        return None

    shard_on_batch = False
    max_cache_length = max_input_tokens + max_output_tokens

    logger.info(f"Creating engine with max_cache_length={max_cache_length} = {max_input_tokens} + {max_output_tokens}")
    env_data = JetEngineEnvironmentData(
        tokenizer_path="", # Tokenizer is not user, HF tokenizer is used instead
        checkpoint_path=model_path,
        checkpoint_format="safetensors",
        batch_size=batch_size,
        max_decode_length=sequence_length,
        max_input_sequence_length=max_input_tokens,
        quant_config=QuantizationConfig(),
        cache_sequence_length=max_cache_length,
        bf16_enable=True,
        sharding_config_path="",
        shard_on_batch=shard_on_batch,
        n_reps=model_info.n_reps,
    )
    env_data.cache_shape = (
        batch_size,
        model_info.num_kv_heads,
        max_cache_length,
        model_info.head_dim,
    )
    env_data.num_layers = model_info.num_layers
    return env_data


def instantiate_model_from_repo_id(
    model_dir: str,
    env: Any,
):
    """Create model instance by hf model_dir, and its config"""
    config = AutoConfig.from_pretrained(model_dir)
    model_info = load_model_info(config)
    # at this point we can be quite optimistic and just assert
    assert model_info is not None

    env.device = "meta"
    model = model_info.model_class.from_config(config, env)
    weights = fetch_models._load_weights(model_dir)
    weights = model.convert_hf_weights(weights)

    model.load_state_dict(weights, assign=True, strict=False)

    return model


def shard_weights(env, weights, weight_shardings):
    """Shard weights according to weight_shardings"""
    for k, v in weight_shardings.items():
        logger.debug(f"SHARDING {k} {v}")
    sharded = {}
    for key, val in weights.items():
        sharding = env.sharding_by_axis(weight_shardings.get(key, -1))
        with jax.default_device(jax.devices("cpu")[0]):
            # Note we clone to avoid a core-dump that might happen otherwise when calling device_put
            arr = torch_xla2.tensor.t2j(val.clone())
        arr = jax.device_put(arr, sharding)
        sharded[key] = torchjax.to_torch(arr)
    return sharded


def create_engine(
    model_path: str,
    batch_size: int,
    sequence_length: int,
    max_input_tokens: int,
    max_output_tokens: int,
) -> PyTorchEngine:
    quantization_enabled = os.environ.get("QUANTIZATION", None) == "jetstream_int8"
    quant_config = QuantizationConfig(enable_weight_quantization=quantization_enabled)
    env_data = create_engine_env_data(model_path, batch_size, sequence_length, max_input_tokens, max_output_tokens)
    if env_data is None:
        return None

    env = JetEngineEnvironment(env_data)
    model = instantiate_model_from_repo_id(model_path, env)
    # NOTE: this is assigned later because, the model should be constructed
    # as a float model first then quantized
    env.quant_config = quant_config
    # Update config with engine data
    model.config.batch_size = batch_size
    model.config.sequence_length = sequence_length

    weight_shardings = model.get_sharding_annotations()
    if quantization_enabled:
        # Some layers are excluded from quantization, to get better accuracy.
        # Get last layer in the model, to exclude it from quantization, guessing the last in the named_modules is the
        # last layer.
        last_layer = list(model.named_modules())[-1][0]
        # Exclude Embedding layers from quantization (in case they are different layers)
        exclude_list = {name for name, module in model.named_modules() if isinstance(module, torch.nn.Embedding)}
        exclude_list.add(last_layer)
        logger.info("Quantizing model to int8")
        start = time.time()
        quant_config = QuantizationConfig(
            enable_weight_quantization=True,
            exclude_layers=exclude_list,
        )
        quantize_model.quantize_model(model, quant_config)
        end = time.time()
        logger.info(f"Quantization took {end - start:.2f} seconds")
    model_weights = model.state_dict()
    sharded_weights = shard_weights(env, model_weights, weight_shardings)
    engine = OptimumJetstreamEngine(
        pt_model=model,
        env=env,
        weights=torchjax.from_torch_with_copy(sharded_weights),
    )
    return engine



#  Copyright 2024 The HuggingFace Team. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any

from transformers import AutoConfig

from optimum.tpu import jetstream_pt_available


def model_can_use_jetstream_pt(model_path: str) -> bool:
    """Checks if the model is supported by Jetstream Pytorch on Optimum TPU and if the required dependencies to provide
    the engine are installed.
    """
    config = AutoConfig.from_pretrained(model_path)
    # For now few models are supported
    supported_models = ["llama", "gemma", "mixtral"]
    if config.model_type not in supported_models:
        return False
    if jetstream_pt_available():
        return True
    return False


def create_engine(
    model_path: str,
    batch_size: int,
    sequence_length: int,
    max_input_tokens: int,
    max_output_tokens: int,
) -> Any:
    if not model_can_use_jetstream_pt(model_path):
        # The model is not compatible with Jetstream PyTorch, just exit
        return None

    # Now import engine_loader to prevent importing it at the top when not supported
    from .engine_loader import create_engine
    return create_engine(
        model_path, batch_size, sequence_length, max_input_tokens, max_output_tokens
    )




import pytest
from decode_tests_utils import DecodeTestParams, decode_single_test


@pytest.mark.parametrize("params",
    [
        DecodeTestParams(
            model_id="google/gemma-2b",
            sequence_length=1024,
            expected_text="\n\nThe first thing I noticed was the smell of the rain. It was a smell I had never",
        ),
        DecodeTestParams(
            model_id="Maykeye/TinyLLama-v0",
            sequence_length=1024,
            expected_text=" It was a very special day, and it was a very special day.\nThe mommy said",
        ),
    ],
    ids=["gemma-2b", "TinyLLama-v0"],
)
def test_decode_single(params):
    decode_single_test(params)

@pytest.mark.slow
@pytest.mark.parametrize("params",
    [
        DecodeTestParams(
            model_id="meta-llama/Meta-Llama-3-8B",
            sequence_length=256,
            expected_text=" Winston Smith, his chin nuzzled into his breast in an effort to escape the vile wind,",
        ),
        DecodeTestParams(
            model_id="google/gemma-7b",
            sequence_length=128,
            expected_text="\n\nThe first line of George Orwell’s <em>1984</em> is a perfect example",
        ),
        DecodeTestParams(
            model_id="mistralai/Mistral-7B-v0.3",
            sequence_length=128,
            expected_text=" Winston Smith, his chin nuzzled into his breast in an effort to escape the v",
        ),
    ],
    ids=["Meta-Llama-3-8B", "gemma-7b", "Mistral-7B-v0.3"],
)
def test_decode_single_slow(params):
    decode_single_test(params)



import asyncio
from pathlib import Path
from typing import List

from grpc import aio
from grpc_reflection.v1alpha import reflection
from loguru import logger

from .auto_generator import AutoGenerator, Generator
from .interceptor import ExceptionInterceptor
from .pb import generate_pb2, generate_pb2_grpc


class TextGenerationService(generate_pb2_grpc.TextGenerationServiceServicer):
    def __init__(self, generator: Generator, server_urls: List[str]):
        self.generator = generator
        self.server_urls = server_urls

    async def Info(self, request, context):
        logger.debug("Info")
        return self.generator.info

    async def Health(self, request, context):
        logger.debug("Health")
        return generate_pb2.HealthResponse()

    async def ServiceDiscovery(self, request, context):
        logger.debug("ServiceDiscovery")
        return generate_pb2.ServiceDiscoveryResponse(urls=self.server_urls)

    async def ClearCache(self, request, context):
        logger.debug("ClearCache")
        if request.HasField("id"):
            self.generator.clear(request.id)
        else:
            self.generator.clear()
        return generate_pb2.ClearCacheResponse()

    async def FilterBatch(self, request, context):
        logger.debug("FilterBatch")
        filtered_batch = self.generator.filter(request.batch_id, request.request_ids)
        return generate_pb2.FilterBatchResponse(batch=filtered_batch)

    async def Warmup(self, request, context):
        logger.info("Warmup (this can take several minutes)")
        max_tokens = self.generator.warmup(request.batch)
        ret = generate_pb2.WarmupResponse(max_supported_total_tokens=max_tokens)
        logger.info("Warmup done")
        return ret

    async def Prefill(self, request, context):
        logger.debug("Prefill")
        batch = request.batch
        generations, batch = self.generator.prefill(request.batch)
        return generate_pb2.PrefillResponse(generations=generations, batch=batch)

    async def Decode(self, request, context):
        logger.debug("Decode")
        generations, batch = self.generator.decode(request.batches)
        return generate_pb2.DecodeResponse(generations=generations, batch=batch)


def serve(
    model_path: str,
    revision: str,
    max_batch_size: int,
    max_sequence_length: int,
    max_input_tokens: int,
    uds_path: Path,
):
    async def serve_inner(model_path: str):
        unix_socket_template = "unix://{}-{}"
        local_url = unix_socket_template.format(uds_path, 0)
        server_urls = [local_url]

        try:
            generator = AutoGenerator.from_pretrained(
                model_path,
                revision=revision,
                max_batch_size=max_batch_size,
                max_sequence_length=max_sequence_length,
                max_input_tokens=max_input_tokens,
            )
        except Exception:
            logger.exception("Error when initializing model")
            raise

        server = aio.server(interceptors=[ExceptionInterceptor()])
        generate_pb2_grpc.add_TextGenerationServiceServicer_to_server(
            TextGenerationService(generator, server_urls), server
        )
        SERVICE_NAMES = (
            generate_pb2.DESCRIPTOR.services_by_name["TextGenerationService"].full_name,
            reflection.SERVICE_NAME,
        )
        reflection.enable_server_reflection(SERVICE_NAMES, server)
        server.add_insecure_port(local_url)

        await server.start()

        logger.info("Server started at {}".format(local_url))

        try:
            await server.wait_for_termination()
        except KeyboardInterrupt:
            logger.info("Signal received. Shutting down")
            await server.stop(0)

    asyncio.run(serve_inner(model_path))



import os

import Levenshtein
import pytest


MODEL_ID = "openai-community/gpt2"
SEQUENCE_LENGTH = 1024


@pytest.fixture(scope="module")
def model_name_or_path():
    os.environ["HF_SEQUENCE_LENGTH"] = str(SEQUENCE_LENGTH)
    yield MODEL_ID


@pytest.fixture(scope="module")
def tgi_service(launcher, model_name_or_path):
    with launcher(model_name_or_path) as tgi_service:
        yield tgi_service


@pytest.fixture(scope="module")
async def tgi_client(tgi_service):
    await tgi_service.health(300)
    return tgi_service.client


@pytest.mark.asyncio
async def test_model_single_request(tgi_client):

    # Greedy bounded without input
    response = await tgi_client.generate(
        "What is Deep Learning?",
        max_new_tokens=17,
        decoder_input_details=True,
    )
    assert response.details.generated_tokens == 17
    assert (
        response.generated_text == "\n\nDeep learning is a technique that allows you to learn something from a set of"
    )

    # Greedy bounded with input
    response = await tgi_client.generate(
        "What is Deep Learning?",
        max_new_tokens=17,
        return_full_text=True,
        decoder_input_details=True,
    )
    assert response.details.generated_tokens == 17
    assert (
        response.generated_text
        == "What is Deep Learning?\n\nDeep learning is a technique that allows you to learn something from a set of"
    )

    # Sampling
    response = await tgi_client.generate(
        "What is Deep Learning?",
        do_sample=True,
        top_k=50,
        top_p=0.9,
        repetition_penalty=1.2,
        max_new_tokens=100,
        seed=42,
        decoder_input_details=True,
    )
    assert (
        'The deep neural networks that we create are essentially "miniature" neural networks that can easily be trained'
        in response.generated_text
    )


@pytest.mark.asyncio
async def test_model_multiple_requests(tgi_client, generate_load):
    num_requests = 4
    responses = await generate_load(
        tgi_client,
        "What is Deep Learning?",
        max_new_tokens=17,
        n=num_requests,
    )

    assert len(responses) == 4
    expected = "\n\nDeep learning is a technique that allows you to learn something from a set of"
    for r in responses:
        assert r.details.generated_tokens == 17
        # Compute the similarity with the expectation using the levenshtein distance
        # We should not have more than two substitutions or additions
        assert Levenshtein.distance(r.generated_text, expected) < 3
